"""
Classes needed for describing a model.

Note:
  For a given model, all parameters, observables and other dependent quantities
  (such as quantities used to impose constraints) must be assigned unique
  string identifiers. Values for parameters and observables are passed between
  functions by means of dictionaries, whose keys are (a subset of) these string
  identifiers.

"""

from math import log, pi, exp
from collections import OrderedDict
from numpy import array, zeros, dot, vdot, eye, diag
from numpy.linalg import inv, det, LinAlgError
import random
import copy
from types import ModuleType


# exceptions

class ParameterError(Exception):
    """Exception indicating invalid parameter values.

    Note:
      ``ParameterError`` should be raised by implementations of
      ``Model.getDependentQuantities`` when the function is called with invalid
      parameters (e.g. nans).

    Args:
      message (str): Error message.
      model (Model instance): Model object whose ``getDependentQuantities``
        method was called.
      parameters (dict): Parameters with which the ``getDependentQuantities``
        method was called.

    """
    def __init__(self, message, model, parameters):
        Exception.__init__(self, message, model, parameters)
        self.message = message
        self.model = model
        self.parameters = parameters


class ComputationError(Exception):
    """Exception indicating a failed computation of dependent quantities.

    Note:
      ``ComputationError`` should be raised by implementations of
      ``Model.getDependentQuantities`` when the function is called with valid
      parameters but the computation fails.

    Args:
      message (str): Error message
      model (Model instance): Model object whose ``getDependentQuantities``
        method was called.
      parameters (dict): Parameters with which the ``getDependentQuantities``
        method was called.

    """
    def __init__(self, message, model, parameters):
        Exception.__init__(self, message, model, parameters)
        self.message = message
        self.model = model
        self.parameters = parameters
    


# helper functions

def _dict_to_matrix(d, keys1, keys2=None):
    if keys2 is None:
        keys2 = keys1
    dim1 = len(keys1)
    dim2 = len(keys2)
    ind1 = dict(zip(keys1, range(dim1)))
    ind2 = dict(zip(keys2, range(dim2)))
    M = zeros((dim1, dim2))
    for (k1, k2), val in d.iteritems():
        M[ind1[k1], ind2[k2]] = val
    return M

def _matrix_to_dict(M, keys1, keys2=None):
    if keys2 is None:
        keys2 = keys1
    dim1 = len(keys1)
    dim2 = len(keys2)
    d = {}
    for i in range(dim1):
        for j in range(dim2):
            if M[i,j] != 0.0:
                d[keys[i], keys[j]] = M[i,j]
    return d

def _dict_to_vector(d, keys):
    dim = len(keys)
    ind = dict(zip(keys, range(dim)))
    v = zeros(dim)
    for i, k in enumerate(keys):
        v[i] = d.get(k, 0.0)
    return v

def _vector_to_dict(v, keys):
    return dict(zip(keys, v))



# parameter class

class Parameter:
    """Class describing a fit parameter.

    Args:
      value (float or None, optional): Value of the parameter. Set to None if
        the parameter is free to float in the fit. Defaults to None.
      default (float, optional): Some sensible default value for the
        parameter. Set to zero if not specified.
      scale (float, optional): The typical size of variations in this parameter.
        Defaults to 1.
      bounds (tuple, optional): Tuple containing the lower and upper bounds (in
        that order) on the parameter. Use None as first or second element to
        leave the parameter unbounded from below or above, respectively.
        Defaults to (None, None).
      generator (callable, optional): Function which generates random values for
        the parameter. The supplied function should take no arguments. If
        omitted or set to ``None``, a sensible default is chosen based on the
        upper and lower bounds. For an unbounded parameter a normal distribution
        with mean `default` and a standard deviation `scale` is chosen. If only
        one bound is present, a log-normal distribution is chosen for the
        distance from that bound. The distribution is centered on `default` and
        has a width of `scale`. If a lower and upper bound are present, the
        distribution is uniform between those bounds.
      rng (object or module, optional): The random number generator to use. This
        is only relevant if `generator` was not specified or set to None. Set it
        to an instance of ``random.Random`` or the ``random`` module itself.
        The latter is the default.

    Attributes:
      value (float or None): Value of the parameter. Set to None if the
        parameter is free to float in the fit. Defaults to None.
      default (float): Default value of the parameter. 
      scale (float): The typical size of variations in this parameter. Defaults
        to 1.
      lowerBound (float or None): Lower bound on the parameter. Set this to None
        to make the parameter unbounded from below. Defaults to None.
      upperBound (float or None): Upper bound on the parameter. Set this to None
        to make the parameter unbounded from above. Defaults to None.
      generate (callable): Function which generates random values for the
        parameter. It takes no arguments. This is either the function supplied
        through the `generator` keyword or the default setting, if no
        `generator` keyword was supplied.

    """

    def generate(self):
        if self.generator is not None:
            return self.generator()
        elif self.lowerBound is None and self.upperBound is None:
            return self._rng.gauss(self.default, self.scale)
        elif self.upperBound is None:
            mu = log(self.default - self.lowerBound)
            sg = self.scale/(self.default - self.lowerBound)
            return self.lowerBound + exp(self._rng.gauss(mu, sg))
        elif self.lowerBound is None:
            mu = log(self.upperBound - self.default)
            sg = self.scale/(self.upperBound - self.default)
            return self.upperBound - exp(self._rng.gauss(mu, sg))
        else:
            return self._rng.uniform(self.lowerBound, self.upperBound)

    def __init__(self, value=None, default=0.0, scale=1.0, bounds=(None,None),
                 generator=None, rng=random):
        self.value = value
        self.default = default
        self.scale = abs(scale)
        self.lowerBound = bounds[0]
        self.upperBound = bounds[1]
        self.generator = generator
        self._rng = rng

    def __deepcopy__(self, memo):
        return Parameter(value=copy.deepcopy(self.value, memo),
                         default=copy.deepcopy(self.default, memo),
                         scale=copy.deepcopy(self.scale, memo),
                         bounds=(copy.deepcopy(self.lowerBound, memo),
                                 copy.deepcopy(self.upperBound, memo)),
                         generator=copy.deepcopy(self.generator, memo),
                         rng=self._rng if isinstance(self._rng, ModuleType) \
                             else copy.deepcopy(self._rng, memo))

    def setBounds(self, l, u):
        """Sets the lower and upper bounds of the parameter.

        Args:
          l (float or None): Lower bound. Use None to leave parameter unbounded
            from below.
          u (float or None): Upper bound. Use None to leave parameter unbounded
            from above.

        """
        self.lowerBound = l
        self.upperBound = u



# constraint classes

class Constraint:
    """Class describing constraints on a model's parameter space.

    Args:
      f (str or callable): The constraint function. The function will be applied
        to the dictionary returned by the ``getDependentQuantities()`` method
        of the constrained model. If `f` is a string, the quantity with that
        name is used as constraint function.
      rel ('<', '>' or '='): The constraint relation.
      bound (float): The bound (right-hand side) of the constraint.
      scale (float, optional): Typical size of variations in the constraint
        function `f`.
      active (bool, optional): If set to ``False`` the constraint is
        ignored. Defaults to ``True``.

    """

    def __init__(self, f, rel, bound, scale=1.0, active=True):
        if rel not in ['<', '>', '=']:
            raise ValueError('Unknown constraint relation `'+rel+'`.')
        scale = abs(scale)
        if rel == '<':
            self._less = True
            self._fact = -1.0/scale
        else:
            self._less = False
            self._fact = +1.0/scale
        if isinstance(f, str):
            self._f = lambda q: q[f]
        else:
            self._f = f
        self.bound = bound
        self._scale = scale
        self.equality = rel == '='
        self.active = active

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, s):
        s = abs(s)
        self._scale = s
        self._fact = -1.0/s if self._less else +1.0/s

    def __call__(self, q):
        """Compute the constraint function.
        
        Args:
          q (dict): Dictionary with all dependent quantities of the model
            to be constrained. This will be the dict returned by the 
            ``getDependentQuantities`` method of the model under consideration.
          
        Returns:
          float: The value `c` of the constraint. If the ``equality`` attribute
            is ``False``, the constraint is satisfied if ``c >= 0``. If the
            ``equaltiy`` attribute is ``True``, constraint is satisfied if
            ``c == 0``.

        """
        return self._fact*(self._f(q) - self.bound)


# chi-square contributions

class ChiSquareContribution:
    """Abstract class describing contributions to the chi-square function.

    Args:
      active (bool, optional): If set to ``False`` the contribution is
      ignored. Defaults to ``True``.

    Attributes:
      active (bool): If set to ``False`` the contribution is ignored. Defaults
      to ``True``.

    """

    def __init__(self, active=True):
        self.active = active

    def __call__(self, q, x):
        """Compute the chi-square contribution.

        Note:
          Derived classes must implement this. The relation between the
          likelihood function `f` and the chi-square function `chisq` 
          (represented by this method) is::

            f(q, x) = exp(-chisq(q,x)/2)

          The integral of `f` over `x` for fixed `q` must be 1. Derived classes
          must ensure that this is the case.

        Args:
          q (dict): Dictionary containing the model parameters and dependent
            quantities, as returned by the model's ``getDependentQuantites``
            method.
          x (dict): Dictionary containing the (observed) values of the 
            observables

        Returns:
          float: The value of the chi-square function for the specified
            parameters `q` and observables `x`.

        """
        raise NotImplementedError

    def min(self, x):
        """Return the smallest possible chi-square value.

        Args:
          x (dict): Dictionary containing the (observed) values of the 
            observables

        """
        raise NotImplementedError

    def getObservableNames(self):
        """Obtain the observables which the chi-square contribution constrains.

        Note:
          Derived classes must implement this.
        
        """
        raise NotImplementedError

    def getCentralValues(self, q):
        """Obtain Minimum of the chi-square contribution in observable space.

        Note:
          Derived classes must implement this.

        Args:
          q (dict): Dictionary holding the model parameters and dependent
            quantites (as returned by ``getDependentQuantites``).

        Returns:
          A dictionary describing the position of the minimum of the chi-square
          contribution in observable space for fixed parameters `q`. Valid keys
          of `x0` are the names of the observables constrained by this
          contribution, as returned by the ``getObservableNames`` method.

        """
        raise NotImplementedError

    def getQuadraticForm(self, q):
        """Obtain the quadratic form on observable space.

        Note:
          Derived classes must implement this.

        Args:
          q (dict): Dictionary holding the model parameters and dependent
            quantites (as returned by ``getDependentQuantites``).

        Returns:
          A dictionary whose entries are of the form ``(i,j) : Qij``, where `i`
          and `j` are names of observables which are constrained by this
          contribution (i.e. elements of the list returned by the
          ``getObservableNames`` method). `Qij` is the corresponding entry in
          the observable space matrix. Only non-zero entries must be included in
          the dictionary. The matrix should be symmetric. For each pair
          ``(i,j)`` with ``i != j`` the pair ``(j,i)`` should also be a valid
          key and the associated value should also be ``Qij``.

        """
        raise NotImplementedError        


class GaussianCSC(ChiSquareContribution):
    """Describes a gaussian contribution with fixed standard deviation.

    Args:
      obs (str): Name of the constrained observable.
      err (float): The standard deviation of the gaussian.
      syst (str or None, optional): Name of the nuisance parameter associated
        with the systematic error. The mean of the gaussian is defined by the
        sum of the quantities `mean` and `syst`. Defaults to ``None``, which
        means that there is no systematic error.
      mean (str, optional): Name of the mean value of `obs`, as it appears in
        the dictionary returned by the model's ``getDependentQuantities``
        method.  Defaults to ``None``, which means that `mean` is the same as
        `obs`.
      active (bool, optional): If set to ``False`` the contribution is
        ignored. Defaults to ``True``.

    Attributes:
      obs (str): Name of the constrained observable.
      err (float): The standard deviation of the gaussian.
      syst (str or None): The name of the nuisance parameter describing the
        systematic error. Defaults to ``None``, which means that there is
        no systematic error.
      mean (str): Name of the mean value of `obs`, as it appears in the
        dictionary returned by the model's ``getDependentQuantities`` method.
        Defaults to `obs`.

    """

    def __init__(self, obs, err, syst=None, mean=None, active=True):
        ChiSquareContribution.__init__(self, active=active)
        self.obs = obs
        self.err = err
        self.mean = obs if mean is None else mean
        self.syst = syst

    def __call__(self, q, x):
        m = q[self.mean] if self.syst is None else q[self.mean] - q[self.syst]
        return ((m - x[self.obs])/self.err)**2 + log(2.0*pi*self.err**2)

    def min(self, x):
        return log(2.0*pi*self.err**2)
    
    def getObservableNames(self):
        return [self.obs]

    def getCentralValues(self, q):
        m = q[self.mean] if self.syst is None else q[self.mean] - q[self.syst]
        return {self.obs:m}

    def getQuadraticForm(self, q):
        return {(self.obs, self.obs) : 1.0/self.err**2}


class CorrelatedGaussianCSC(ChiSquareContribution):
    """Describes the contribution of several correlated gaussian observables.

    Args:
      sigma (dict): Dictionary containing the standard deviations of all
        contributing observables.
      cor (dict): Dictionary containing the entries of the correlation matrix.
        The keys in `cor` must be tuples (`o1`, `o2`), where `o1` and `o2`
        must be valid keys of `sigma`. If (`o1`, `o2`) appears in `cor`,
        (`o2`, `o1`) must not appear. For pairs not appearing in `cor`, a
        correlation coefficient of 0 is assumed.
      means (dict, optional): Dictionary containing the names of the mean values
        for the constrained observables. `means` must contain key-value pairs
        `o`:`m`, where `o` is a valid key in `sigma` and `m` is a valid key in
        the dictionary returned by the model's ``getDependentQuantities``
        method.
      active (bool, optional): If set to ``False`` the contribution is
        ignored. Defaults to ``True``.

    Raises:
      ValueError: If the correlations specified in `cor` are invalid.

    """
    
    def __init__(self, sigma, cor, means={}, active=True):
        ChiSquareContribution.__init__(self, active=active)
        self.obs, err = zip(*sigma.items())
        err = array(err)
        inverr = array([1.0/e for e in err])
        self.means = [means.get(o, o) for o in self.obs]
        self.dim = len(self.means)
        ind = dict(zip(self.obs, range(self.dim)))
        cov = eye(self.dim)
        for (o1, o2), c in cor.iteritems():
            if (o2, o1) in cor:
                raise ValueError(
                    'Double appearence of ('+o1+', '+o2+') correlation.')
            if o1 == o2:
                raise ValueError(
                    'Diagonal entry for '+o1+' encountered.')
            i = ind[o1]
            j = ind[o2]
            cov[i,j] = cov[j,i] = c
        self.detvar = det(cov)*(err**2).prod()
        if self.detvar <= 0.0:
            raise ValueError('Non-positive variance matrix encountered.')
        try:
            self.invvar = dot(dot(diag(inverr), inv(cov)), diag(inverr))
        except LinAlgError:
            raise ValueError('Singular variance matrix encountered.')
                                                 
    def __call__(self, q, x):
        try:
            v  = array([x[o] for o in self.obs])
            v0 = array([q[m] for m in self.means])
        except KeyError:
            raise KeyError('Missing observable or mean value.')
        dv = v - v0
        return vdot(dv, dot(self.invvar, dv)) + \
            self.dim*log(2.0*pi) + log(self.detvar)

    def min(self, x):
        return self.dim*log(2.0*pi) + log(self.detvar)

    def getObservableNames(self):
        return self.obs[:]
                                                 
    def getCentralValues(self, q):
        return dict((o, q[m]) for o, m in zip(self.obs, self.means))
                                                 
    def getQuadraticForm(self, q):
        result = {}
        for i in range(len(self.obs)):
            for j in range(len(self.obs)):
                result[self.obs[i],self.obs[j]] = self.invvar[i,j]
        return result
                                                 


# model class

class Model:
    """Base class for describing a model.

    Attributes:
      par (dict): Specifies the independent parameters of the model. The keys
       must be strings. The values must be ``Parameter`` instances.
      obs (dict): Specifies the 'measured' values of the observables. The keys
        must be strings. The values must be floats.
      con (dict): Contains the constraints imposed on the model's parameter
        space. The keys must be strings. The values must be instances of
        `Constraint`.
      csc (dict): Contains the contributions of the chi-square function. The
        keys must be strings. The values must be instances of
        `ChiSquareContribution`.
      nuisanceCSCs (set of str): Names of the CSCs which are associated to
        nuisance parameters. 

    """

    def __init__(self):
        self.par = OrderedDict([])
        self.obs = OrderedDict([])
        self.con = {}
        self.csc = {}
        self.nuisanceCSCs = set()


    def copy(self):
        """
        Return a deep copy of the model.

        """
        return copy.deepcopy(self)

    # adding parameters, chi-square contributions and constraints

    def addParameter(self, name, **kwargs):
        """Add a parameter to the model.

        Args:
          name (str): Unique string which identifies the parameter.

          \\*\\*kwargs (optional): Keyword args to be passed to the constructor
            of the ``Parameter`` class.

        """
        if name in self.par:
            raise NameError('Parameter `'+name+'` already exists.')
        self.par[name] = Parameter(**kwargs)

    def addSystematicNP(self, name, val, syst):
        """Add a nuisance parameter with a systematic error to the model.

        Args:
          name (str): Name of the nuisance parameter. This is the key under
            which the parameter is added to the ``par`` dict.
          val (float): Central value of the parameter. This value serves
            as default value for the parameter.
          syst (float or tuple of floats): Systematic error. If a float is
            given, a symmetric systematic error is assumed. For a tuple, the
            first element is the 'plus' error and the second element the
            'minus' error. Both numbers should be positive.
        """
        (systp, systm) = (syst, syst) if not hasattr(syst, '__len__') else syst
        self.addParameter(name, default=val, scale=0.5*(systp+systm),
                          bounds=(val-systm, val+systp))

    def addGaussianNP(self, name, val, err, syst=None):
        """Add a nuisance parameter with a gaussian error to the model.

        This adds a properly initialised ``Parameter`` instance to the
        ``par`` dict and a properly initialised ``GaussianCSC`` instance
        to the ``csc`` dict.

        Args:
          name (str): Name of the nuisance parameter. This is the key under
            which the parameter is added to the ``par`` dict and also the key
            under which the GaussianCSC instance is added to the ``csc`` dict.
          val (float): Central value of the parameter. This value serves
            as default value for the parameter and as mean value for
            the gaussian chi-square contribution.
          err (float): The standard deviation of the gaussian. This value
            serves as scale for the parameter and as standard deviation for
            the gaussian chi-square contribution.
          syst (None or float or tuple of floats, optional): Systematic
            error. See documentation of ``addGaussianCSC``.

        """
        self.addParameter(name, default=val, scale=err)
        if self.obs.get(name, val) != val:
            raise NameError('Observable `'+name+'` already exists.')
        self.obs[name] = val
        self.addGaussianCSC(name, val, err, syst=syst)
        self.nuisanceCSCs.add(name)

    def addCorrelatedGaussianNP(self, name, par, cor):
        """Add a correlated Gaussian nuisance parameters to the model.

        This adds several properly initialised ``Parameter`` instances to the
        ``par`` dict and a properly initialised ``CorrelatedGaussianCST``
        instance to the ``csc`` dict.

        Args:
          name (str): Name identifying the set of nuisance parameters. The
            ``CorrelatedGaussianCSC`` instance is added to the ``csc`` dict
            under this key.

          par (iterable): Contains tuples of the form ``(parname, val, err)``
            where `parname` is the name of a (new) nuisance parameter, `val` is
            the mean value of that parameter `err` the standard deviation in the
            chi-square contribution. `val` will also be used as the parameters
            default value and `err` as its scale.
          cor (iterable): Contains tuple of the form ``(par1, par2, r)``
            where `par1` and `par2` are parameter names previously mentioned
            in the `par` argument and `r` is the associated correlation
            coefficient.

        """
        for n, v, e in par:
            self.addParameter(n, default=v, scale=e)
            if self.obs.get(n, v) != v:
                raise NameError('Observable `'+n+'` already exists.')
            self.obs[n] = v
        self.addCorrelatedGaussianCSC(name, par, cor)
        self.nuisanceCSCs.add(name)

    def addGaussianCSC(self, name, val, err, syst=None, **kwargs):
        """Add a gaussian contribution to the chi-square function.

        Args:
          name (str): Name of the observable. It should be a valid key
            in the dictionary returned by ``getDependentQuantities``. The
            associated value defines the mean of the gaussian distribution.
            `name` is also the key under which an instance of ``GaussianCSC``
            is added to the ``csc`` dict.
          val (float): Measured value of the observable. This value is added
            to the ``obs`` dict under the key `name`.
          err (float): The standard deviation of the gaussian.

          syst (None or float or tuple of floats, optional): Systematic
            error(s).  If the value is not ``None`` a *nuisance parameter* is
            added to the model. The name of this parameter is composed by
            appending ' syst' to `name` and its range is from ``-syst`` to
            ``+syst`` if `syst` is a float or from ``-syst[1]`` to ``+syst[0]``
            if `syst` is a tuple. The mean of the gaussian is then defined by
            the sum of the quantity `name` and the nuisance parameter.  Defaults
            to ``None``, which means that there is no systematic error.
          \*\*kwargs (optional): Keyword args to be passed to the constructor
            of the ``GaussianCSC`` class.

        Raises:
          NameError: Raised when `name` is already used in the ``csc``
            or ``obs`` dicts.

        """
        if name in self.csc:
            raise NameError(\
                'Chi-square contribution `'+name+'` already exists.')
        if self.obs.get(name, val) != val:
            raise NameError('Observable `'+name+'` already exists.')
        self.obs[name] = val
        if syst is not None:
            (systp, systm) = (syst, syst) if not hasattr(syst , '__len__') \
                else syst
            self.addParameter(name+' syst',
                              default = 0.0,
                              scale = 0.5*(systp + systm),
                              bounds = (-systm, systp))
            systpar = name+' syst'
        else:
            systpar = None
        self.csc[name] = GaussianCSC(name, err, syst=systpar, **kwargs)
            
    def addCorrelatedGaussianCSC(self, name, obs, cor, **kwargs):
        """Add a correlated Gaussian contribution to the chi-square function.

        Args:
          name (str): Name identifying the contribution. A
            ``CorrelatedGaussianCSC`` instance is added to the ``csc`` dict
            under this key.
          obs (iterable): Contains tuples of the form ``(obsname, val, err)``
            where `obsname` is the name of an observable, `val` is the measured
            value and `err` the standard deviation. `obsname` should be a valid
            key in the dictionary returned by ``getDependentQuantities``. The
            associated value defines the mean of the gaussian distribution.
            The value `val` is added to the ``obs`` dict under the key
            `obsname`.
          cor (iterable): Contains tuple of the form ``(obs1, obs2, r)``
            where `obs1` and `obs2` are observable names previously mentioned
            in the `obs` argument and `r` is the associated correlation
            coefficient.

        Raises:
          NameError: Raised when `name` is already used in the ``csc``
            dict or any of the observable names in the `obs` argument are
            mentioned twice or already exist as keys in the ``obs`` attribute.
            It is also raised when a correlation coefficient is mentioned twice
            in the `cor` argument.

        """
        if name in self.csc:
            raise NameError(\
                'Chi-square contribution `'+name+'` already exists.')
        errdict = {}
        for n, v, e in obs:
            if self.obs.get(n, v) != v:
                raise NameError('Observable `'+n+'` already exists.')
            self.obs[n] = v
            errdict[n] = e
        cordict = {}
        for n1, n2, c in cor:
            if (n1, n2) in cordict or (n2, n1) in cordict:
                raise NameError(\
                    'Double mention of correlation for observables `'+\
                        n1+'` and `'+n2+'`.')
            cordict[n1, n2] = c
        self.csc[name] = CorrelatedGaussianCSC(errdict, cordict, **kwargs)

    def addConstraint(self, f, rel, bound, scale=1.0, active=True, name=None):
        """Add a constraint on a model's parameter space.

        Args:

          f (str or callable): The constraint function. The function will be
            applied to the dictionary returned by the
            ``getDependentQuantities()`` method of the constrained model. If `f`
            is a string, the dependent quantity with that name is used as
            constraint function.
          rel (``'<'``, ``'>'``, ``'='`` or ``'in'``): The constraint relation.
            With ``'in'`` you can constraint `f` to lie in a finite range. In
            this case two constraints (one for the lower and one for the upper
            bound) are added to the :py:attr:`~Model.con` dictionary.
          bound (float): The bound (right-hand side) of the constraint. If
            `rel` is ``'in'`` this must be a tuple holding the lower and upper
            bound.
          scale (float, optional): Typical size of variations in the constraint
            function `f`.
          active (bool, optional): If set to ``False`` the constraint is
            ignored. Defaults to ``True``.
          name (str, optional): The name of the constraint (i.e. the key under
            which it is added to the :py:attr:`~Model.con` dictionary). If `f`
            is a callable this must be specified. If `f` is a string and `name`
            is not specified a name is generated by appending `rel` to `f`. If
            `rel` is ``'in'`` the value of `name` can be a tuple of two strings
            (the name for the lower bound and the name for the upper bound) or a
            string (in which case names for the lower and upper bounds are
            generated by appending ``'<'`` and ``'>'`` to `name`).

        """
        if rel not in ['<', '>', '=', 'in']:
            raise ValueError('Unknown constraint relation `'+rel+'`.')
        if name is None:
            if isinstance(f, str):
                name = f
            else:
                raise ValueError('Missing name for constraint relation.')
        if rel == 'in':
            if hasattr(bound, '__len__') and len(bound) == 2:
                if isinstance(name, str):
                    name1 = name + ' >'
                    name2 = name + ' <'
                else:
                    name1 = name[0]
                    name2 = name[1]
                self.addConstraint(f, '>', bound[0], 
                                   scale=scale, active=active, name=name1)
                self.addConstraint(f, '<', bound[1],
                                   scale=scale, active=active, name=name2)
            else:
                raise ValueError('To use `in` relation `bound` argument must be a tuple.')
        else:
            self.con[name] = Constraint(f, rel, bound,
                                        scale=scale, active=active)


    # fixing/releasing parameters

    def fix(self, *args):
        """Fix model parameters.

        Args:
          *args (str or list of str or dict): Parameters to fix. For string
            arguments or lists of strings the
            corresponding parameter(s) are fixed to their default values.

        """
        for arg in args:
            if hasattr(arg, 'keys'):
                for name, val in arg.iteritems():
                    self.par[name].value = val
            elif isinstance(arg, str):
                self.par[arg].value = self.par[arg].default                
            elif hasattr(arg, '__iter__'):
                for a in arg:
                    self.fix(a)
            else:
                raise ValueError('Arguments must be dicts, strings or iterables.')

    def release(self, *args):
        """Release model parameters (allowing them to float in a fit).

        Args:
          *args (str or list of str): Names of the parameters to release

        """
        for arg in args:
            if isinstance(arg, str):
                self.par[arg].value = None
            elif hasattr(arg, '__iter__'):
                for p in arg:
                    self.release(p)
            else:
                raise ValueError('Arguments must be strings or iterables.')


    # activating/deactivating chi-square contributions

    def activateAllCSCs(self, ignorenp=True):
        """Activate all chi-square contributions.

        Args:
          ignorenp (bool, optional): If ``True``, settings for chi-square
            contributions from nuisance parameters are not altered. Defaults to
            ``True``.

        """
        for n, c in self.csc.iteritems():
            if not (ignorenp and n in self.nuisanceCSCs):
                c.active = True

    def deactivateAllCSCs(self, ignorenp=True):
        """Deactivate all chi-square contributions.

        Args:
          ignorenp (bool, optional): If ``True``, settings for chi-square
            contributions from nuisance parameters are not altered. Defaults to
            ``True``.

        """
        for n, c in self.csc.iteritems():
            if not (ignorenp and n in self.nuisanceCSCs):
                c.active = False

    def activateCSCs(self, *args):
        """Activate given set of chi-square contributions.

        Args:
          *args (str): Names of the chi-square contributions to activate.

        """
        for name in args:
            self.csc[name].active = True

    def deactivateCSCs(self, *args):
        """Deactivate given set of chi-square contributions.

        Args:
          *args (str): Names of the chi-square contributions to deactivate.

        """
        for name in args:
            self.csc[name].active = False

    def activateOnlyCSCs(self, *args, **kwargs):
        """Activate given set of chi-square contributions, deactivate all others.

        Args:
          *args (str): Names of the chi-square contributions to activate.
          ignorenp (bool, optional): If ``True``, settings for chi-square
            contributions from nuisance parameters are not altered. Defaults to
            ``True``.

        """
        self.deactivateAllCSCs(**kwargs)
        self.activateCSCs(*args)

    def deactivateOnlyCSCs(self, *args, **kwargs):
        """Deactivate given set of chi-square contributions, activate all others.

        Args:
          *args (str): Names of the chi-square contributions to deactivate.
          ignorenp (bool, optional): If ``True``, settings for chi-square
            contributions from nuisance parameters are not altered. Defaults to
            ``True``.

        """
        self.activateAllCSCs(**kwargs)
        self.deactivateCSCs(*args)


    # activating/deactivating constraints

    def activateAllConstraints(self):
        """Activate all constraints.

        """
        for n, c in self.con.iteritems():
            c.active = True

    def deactivateAllConstraints(self):
        """Deactivate all constraints.

        """
        for n, c in self.con.iteritems():
            c.active = False

    def activateConstraints(self, *args):
        """Activate given set of constraints.

        """
        for name in args:
            self.con[name].active = True

    def deactivateConstraints(self, *args):
        """
        Activate given set of chi-square contributions.

        """
        for name in args:
            self.con[name].active = False

    def activateOnlyConstraints(self, *args):
        """Activate given set of constraints, deactivate all others.

        """
        self.deactivateAllConstraints()
        self.activateConstraints(*args)

    def deactivateOnlyConstraints(self, *args):
        """Deactivate given set of constraints, activate all others.

        """
        self.activateAllConstraints()
        self.deactivateConstraints(*args)


    # get information about observables, parameters and constraints

    def getObservableNames(self):
        """Obtain the names of the observables predicted by the model.

        Returns:
          A list of strings containing the names of the observables predicted by
          the model.

        Note:
          This list is compiled by calling the ``getObservableNames`` method of
          all elements of the `csc` attribute and has nothing to do with the
          keys found in the `obs` attribute.

        """
        r = []
        for name, c in self.csc.iteritems():
            if c.active:
                obs = c.getObservableNames()
                for o in obs:
                    if o in r:
                        raise NameError('Multiple chi-square contributions for observable `'+o+'`.')
                r.extend(obs)
        return r

    def getFreeParameters(self):
        """Obtain the names of the parameters floating in the fit.

        Returns:
          A list of strings containing the names of the parameters which are
          free to float in the fit (possibly with lower or upper bounds).

        """
        return [name for name, p in self.par.iteritems() if p.value is None]

    def getFixedParameters(self):
        """Obtain a dictionary of the fixed parameters and their values.

        Returns:
          A dictionary containing the names and values of the parameters which
          are fixed.

        """
        return dict((name, p.value) for name, p in self.par.iteritems() \
                        if p.value is not None)

    def getDefaultParameters(self):
        """Obtain a dictionary with the default values of all model parameters.

        """
        return dict((name, p.default) for name, p in self.par.iteritems())


    # computation methods

    def getDependentQuantities(self, p):
        """Compute all quantities depending on the model parameters.

        Args:
          p (dict): Contains the values of the model parameters. The keys must
            be the same as those of the `par` member. The values must be
            floats.

        Returns:
          A dictionary containing the parameters from `p` and all quantities
          which depend on them. The keys are strings and the values are floats.

        Note:
          The default implementation returns a dictionary with the full set of
          model parameters, using default values for parameters not specified in
          `p` and overriding the values of fixed parameters. Derived classes
          should add all dependent quantities to the dictionary.  Instances of
          ``ChiSquareContribution`` and ``Constraint`` will use the dictionary
          returned by this method to compute the chi-square contributions and
          the values of the constraint functions, so all quantities needed for
          these computations should be included.  For instances of
          ``GaussianCSC`` this means including the mean value of the
          corresponding observable. You may use the same string identifier for
          the observable and the mean.

        """
        res = self.getDefaultParameters()
        res.update(p)
        res.update(self.getFixedParameters())
        return res

    def getConstraints(self, q):
        """Compute the constraint functions.

        Args:
          q (dict): Contains the parameters and dependent quantities, as
            returned by the ``getDependentQuantities`` method.

        Returns:
          A dictionary containing the values of the active constraint
          functions. The keys are a subset of those found in the `con`
          member. The values are floats.

        """
        return dict((name, c(q)) for name, c in self.con.iteritems() \
                        if c.active)

    def getConstraintViolation(self, q):
        """Compute the amount of constraint violation.

        Note:
          This function returns the sum of the squares of the constraint
          functions for all violated constraints

        Args:
          q (dict): Contains the parameters and dependent quantities, as
            returned by the ``getDependentQuantities`` method.

        """
        c = self.getConstraints(q)
        return sum(val**2 for name, val in c.iteritems() \
                       if self.con[name].equality or val < 0.0)

    def getChiSquare(self, q, obs=None):
        """Compute the chi-square value from dependent quantities.

        This function computes the chi-square value by calling the
        `getChiSquare` methods of all active ``ChiSquareContribution`` objects
        in the `csc` member and adding up the results.

        Args:
          q (dict): Contains the parameters and dependent quantities, as
            returned by the ``getDependentQuantities`` method.
          obs (dict, optional): Contains the measured values of the observables.
            The keys must be those returned by `getObservableNames`. The values
            must be floats. If omitted, the content of the `obs` member is used.

        Returns:
          float: The value of the chi-square function.

        """
        if obs is None:
            obs = self.obs
        return sum(c(q, obs) for name, c in self.csc.iteritems() if c.active)

    def getDeltaChiSquare(self, q, obs=None):
        """Compute the Delta chi-square value from dependent quantities.

        This function computes the chi-square value minus the absolute
        lower bound on the chi-square value by calling the
        `getChiSquare` and `min` methods of all active
        ``ChiSquareContribution`` objects in the `csc` member and adding up the
        results.

        Args:
          q (dict): Contains the parameters and dependent quantities, as
            returned by the ``getDependentQuantities`` method.
          obs (dict, optional): Contains the measured values of the observables.
            The keys must be those returned by `getObservableNames`. The values
            must be floats. If omitted, the content of the `obs` member is used.

        Returns:
          float: The value of the chi-square function.

        """
        if obs is None:
            obs = self.obs
        return sum(c(q, obs) - c.min(obs) \
                       for name, c in self.csc.iteritems() if c.active)

    def chisq(self, p, obs=None, only=None, exclude=[]):
        """Compute the chi-square value for given parameters and observables.

        Args:
          p (dict): Contains the values of the free parameters
          obs (dict, optional): Contains the values of the observables. Uses
            self.obs if omitted.
          only (list or None, optional): Only include the listed chi-square
            contributions.  Defaults to ``None``, in which case all active
            chi-square contributions are included. Inactive contributions are
            never included.
          exclude (list, optional): Exclude listed chi-square contributions from
            sum.  Defaults to ``[]``.

        Returns:
          float: The chi-square value for parameters p and observables obs.

        """
        if obs is None:
            obs = self.obs
        q = self.getDependentQuantities(p)
        return sum(c(q, obs) for name, c in self.csc.iteritems() \
                       if c.active and (only is None or name in only) and \
                       name not in exclude)

    def dchisq(self, p, obs=None, only=None, exclude=[]):
        """Compute the Delta chi-square value for given parameters and observables.

        Args:
          p (dict): Contains the values of the free parameters
          obs (dict, optional): Contains the values of the observables. Uses
            self.obs if omitted.
          only (list or None, optional): Only include the listed chi-square
            contributions.  Defaults to ``None``, in which case all active
            chi-square contributions are included. Inactive contributions are
            never included.
          exclude (list, optional): Exclude listed chi-square contributions from
            sum.  Defaults to ``[]``.

        Returns:
          float: The chi-square value for parameters p and observables obs.

        """
        return self.chisq(p, obs=obs, only=only, exclude=exclude) - \
            self.minChisq(obs=obs, only=only, exclude=exclude)

    def minChisq(self, obs=None, only=None, exclude=[]):
        if obs is None:
            obs = self.obs
        return sum(c.min(obs) for name, c in self.csc.iteritems() \
                       if c.active and (only is None or name in only) and \
                       name not in exclude)
        
    def getCentralValues(self, q):
        """Obtain Minimum of the chi-square function in observable space.

        Args:
          q (dict): Dictionary holding the model parameters and dependent
            quantites (as returned by ``getDependentQuantites``).

        Returns:
          A dictionary describing the position of the minimum of the chi-square
          function in observable space for fixed parameters `q`. Valid keys are
          the names of the observables constrained by the chi-square function,
          as returned by the ``getObservableNames`` method.

        """
        obs = self.getObservableNames()
        x0 = {}
        for name, c in self.csc.iteritems():
            if c.active:
                x0.update(c.getCentralValues(q))
        return x0

    def generateParameterPoint(self):
        """Generate a random point in parameter space.

        Returns:
          A dictionary of the randomly generated parameters. Fixed parameters
          are included in the dictionary and will have their fixed values.

        """
        return dict((name, p.generate() if p.value is None else p.value) \
                        for name, p in self.par.iteritems())

    def getQuadraticForm(self, q):
        """Obtain the quadratic form on observable space.

        Note:
          The quadratic form is obtained by calling the ``getQuadraticForm``
          method of all active chi-square contributions and adding up the
          results.

        Args:
          q (dict): Dictionary holding the model parameters and dependent
            quantites (as returned by ``getDependentQuantites``).

        Returns:
          A dictionary whose entries are are of the form ``(i,j) : Qij``, where
          `i` and `j` are names of observables which are constrained by the
          chi-square function (i.e. elements of the list returned by the
          ``getObservableNames`` method). `Qij` is the corresponding entry in
          the observable space matrix. Zero entries may be omitted in the
          dictionary. The matrix is symmetric.

        """
        result = {}
        for name, c in self.csc.iteritems():
            if c.active:
                for ij, v in c.getQuadraticForm(q).iteritems():
                    result[ij] = result.get(ij, 0.0) + v
        return result


