import numpy
import copy
from histograms import Histogram1D, Histogram2D

class Profiler2D:
    """Compute the profile likelihood on a two-dimensional space.

    Note:
      When the profile likelihood function is to be computed on a 2D grid the
      stability of the optimisations can be significantly improved by using
      the best-fit parameters of neighbouring grid points as starting points for
      local optimisations. The :py:class:`~myFitter.Profiler1D` and
      :py:class:`~myFitter.Profiler2D` classes implement this method for one and
      two-dimensional scans.

      For difficult optimisation problems you may need rather detailed control
      over the procedure for finding the global minima. In this case you can
      derive a class from Profile2D and override the
      :py:meth:`~myFitter.Profiler2D.fit` method.
      
      The two variables in which the profile likelihood is computed can be
      arbitrary dependent quantities (not only model parameters). For dependent
      quantities which are not model parameters a constraint is (temporarily)
      added to the model. The names of these constraints are ``'__xscan__'``
      and ``'__yscan__'``. You should therefore avoid using these names for your
      own constraints.

    Args:
      xvar (str): Name of the variable on the x-axis. Must
        be the name of a dependent quantitiy in the model (not necessarily
        a parameter).
      yvar (str): Name of the variable on the y-axis. Must
        be the name of a dependent quantitiy in the model (not necessarily
        a parameter).
      fitter (Fitter or None, optional): The Fitter object to use for global
        minimisations. The model whose likelihood is profiled will be
        ``fitter.model``. Defaults to ``None``, in which case the `model`
        argument must be given.
      model (Model or None, optional): The model whose likelihood function
        shall be profiled. By specifying the `model` keyword instead of `fitter`
        you disable global optimisations by Fitter objects. Use this only if 
        local optimisations reliably find the global minimum (e.g. because your
        likelihood function is convex). Defaults to ``None``, in which case
        `fitter` must be specified.
      xscale (float, optional): Scale for the constraint associated
        with the variable on the x-axis. (See :py:class:`~myFitter.Constraint`
        for details.) This argument is irrelevant if the corresponding variable
        is a model parameter.
      yscale (float, optional): Scale for the constraint associated
        with the variable on the y-axis. (See :py:class:`~myFitter.Constraint`
        for details.) This argument is irrelevant if the corresponding variable
        is a model parameter.
      nsamples (int, optional): Number of points to sample for each global
        minimisation. Defaults to 100.
      tries (int, optional): Number of local minimisations to try for finding
        the global minimum. Defaults to 5.
      log (output stream or None, optional): Output stream for log messages.
        Defaults to ``None``, in which case no messages are written.
      useneighbors (bool, optional): Whether to use best-fit parameters of
        neighbouring cells as starting points. Defaults to ``True``.

    """
    def __init__(self, xvar, yvar, fitter=None, model=None,
                 xscale=1.0, yscale=1.0, nsamples=100, tries=5, log=None,
                 useneighbors=True):
        self.xvar = xvar
        self.yvar = yvar
        self.fitter = fitter
        if model is None:
            if fitter is not None:
                self.model = fitter.model
            else:
                raise ValueError('Specify either `fitter` or `model` keyword.')
        else:
            if fitter is not None:
                raise ValueError('Specify either `fitter` or `model` keyword.')
            else:
                self.model = model
        self._xcons = self.xvar not in self.model.par.keys()
        self._ycons = self.yvar not in self.model.par.keys()
        self._xscale = xscale
        self._yscale = yscale

        self.nsamples = nsamples
        self.tries = tries
        self.log = log
        self.useneighbors = useneighbors

    def fix(self, x, y):
        """Fix the variables of the profile likelihood.

        Note:
          If the variables of the profile likelihood are model parameters they
          are fixed with :py:meth:`myFitter.Model.fix`. Otherwise the ``bound``
          members of the corresponding constraints are adjusted. You will
          typically use this function in a derived class when you override the
          :py:meth:`myFitter.Profiler2D.fit` method.

        Args:
          x, y (float): Values of the x and y variables.

        """
        if self._xcons:
            self.model.con['__xscan__'].bound = x
        else:
            self.model.fix({self.xvar : x})
        if self._ycons:
            self.model.con['__yscan__'].bound = y
        else:
            self.model.fix({self.yvar : y})

    def fit(self, xval=None, yval=None, start=[], sample=True,
            nsamples=100, tries=5, log=None, **kwargs):
        """Minimise the chi-square for a single grid point.

        Note:
          Derived classes can override this function to implement specific
          methods for global optimisation. The function should fix the
          scan variables to `xval` and `yval`, respectively (using
          :py:meth:`~myFitter.Profiler2D.fix`), do local optimisations with all
          the starting points listed in `start`, and perform an additional global
          optimisation if and only if `sample` is ``True``. The default behaviour
          is to randomly sample a number of points set by the
          :py:attr:`~myFitter.Profiler2D.nsamples` attribute and then try
          a number of local optimisations set by
          :py:attr:`~myFitter.Profiler2D.tries`.

        Args:
          xval, yval (float): Values of the x and y variables.
          start (list of dicts, optional): Starting points for local
            optimisations. The dictionaries will hold values for the model
            parameters. Defaults to ``[]``.
          sample (bool, optional): Indicates if an additional global
            optimisation (i.e. random sampling followed by several local
            optimisations) is required. Defaults to ``True``.
          nsamples (int, optional): Number of points to sample for each global
            minimisation. Defaults to 100.
          tries (int, optional): Number of local minimisations to try for finding
            the global minimum. Defaults to 5.
          log (output stream or None, optional): Output stream for log messages.
            Defaults to ``None``, in which case no messages are written.
          **kwargs (optional): additional keyword arguments. The default
            implementation passes them to the :py:meth:`myFitter.Fitter.fit`
            method.

        """
        if log is not None:
            if sample:
                log.write('pass 1, x='+repr(xval)+', y='+repr(yval)+'\n')
            else:
                log.write('pass 2, x='+repr(xval)+', y='+repr(yval)+'\n')
            log.flush()
        self.fix(xval, yval)
        if sample:
            self.fitter.clear()
            self.fitter.sample(nsamples)
        else:
            tries = 0
        return self.fitter.fit(tries=tries, start=start, **kwargs)

    def scan(self, hist, **kwargs):
        """Compute the profile likelihood on a specified grid.

        Args:
          hist (Histogram2D): A 2D histogram whose
            :py:attr:`~myFitter.Histogram2D.xvals`
            and :py:attr:`~myFitter.Histogram2D.yvals` members specify the grid
            on which the profile likelihood is to be computed. After a
            successful scan the data values of `hist` will be
            :py:class:`~myFitter.FitResult` objects
            holding information about the fit which yielded the smallest
            chi-square value.
          nsamples (int, optional): Number of points to sample for each global
            minimisation. Default value set by constructor.
          tries (int, optional): Number of local minimisations to try for finding
            the global minimum. Defaults value set by constructor.
          log (output stream or None, optional): Output stream for log messages.
            If ``None`` no messages are written. Default value set in
            constructor.
          useneighbors (bool or None, optional): Whether to use best-fit
            parameters of neighbouring cells as starting points. Default
            value is set in constructor.
          **kwargs (optional): Any additional keyword arguments are passed to
            the :py:meth:`~myFitter.Profiler2D.fit` method.

        """
        if 'nsamples' not in kwargs:
            kwargs['nsamples'] = self.nsamples
        if 'tries' not in kwargs:
            kwargs['tries'] = self.tries
        if 'log' not in kwargs:
            kwargs['log'] = self.log

        useneighbors = kwargs.pop('useneighbors', self.useneighbors)

        par = self.model.par
        self.model.par = copy.deepcopy(par)
        if self._xcons:
            self.model.addConstraint(self.xvar, '=', 0.0,
                                     scale=self._xscale, 
                                     active=True,
                                     name='__xscan__')
        if self._ycons:
            self.model.addConstraint(self.yvar, '=', 0.0,
                                     scale=self._yscale,
                                     active=True,
                                     name='__yscan__')

        if useneighbors:
            for k in range((len(hist.yvals) + 1) // 2):
                j = 2*k
                y = hist.yvals[j]
                for i, x in enumerate(hist.xvals):
                    sp = []
                    if j > 0:
                        sp.append(hist.data[j-1][i].bestFitParameters)
                    if i > 0:
                        sp.append(hist.data[j][i-1].bestFitParameters)
                    r = self.fit(xval=x, yval=y, start=sp, sample=True,
                                 **kwargs)
                    hist.data[j][i] = min(hist.data[j][i], r)


                j = 2*k+1
                if j < len(hist.yvals):
                    y = hist.yvals[j]
                    if len(hist.xvals) > 1:
                        for i, x in reversed(list(enumerate(hist.xvals))):
                            sp = []
                            if j > 0:
                                sp.append(hist.data[j-1,i].bestFitParameters)
                            if i < len(hist.xvals)-1:
                                sp.append(hist.data[j,i+1].bestFitParameters)
                            r = self.fit(xval=x, yval=y, start=sp, sample=True,
                                         **kwargs)
                            hist.data[j][i] = min(hist.data[j][i], r)

            if len(hist.yvals) > 1:
                for k in reversed(range((len(hist.yvals) + 1) // 2)):
                    j = 2*k+1
                    if j < len(hist.yvals):
                        y = hist.yvals[j]
                        for i, x in enumerate(hist.xvals):
                            sp = []
                            if j < len(hist.yvals)-1:
                                sp.append(hist.data[j+1,i].bestFitParameters)
                            if i > 0:
                                sp.append(hist.data[j,i-1].bestFitParameters)
                            if sp:
                                r = self.fit(xval=x, yval=y, start=sp,
                                             sample=False, **kwargs)
                                hist.data[j][i] = min(hist.data[j][i], r)

                    j = 2*k
                    y = hist.yvals[j]
                    if len(hist.xvals) > 1:
                        for i, x in reversed(list(enumerate(hist.xvals))):
                            sp = []
                            if j < len(hist.yvals)-1:
                                sp.append(hist.data[j+1][i].bestFitParameters)
                            if i < len(hist.xvals)-1:
                                sp.append(hist.data[j][i+1].bestFitParameters)
                            if sp:
                                r = self.fit(xval=x, yval=y, start=sp,
                                             sample=False, **kwargs)
                                hist.data[j][i] = min(hist.data[j][i], r)
        else:
            for j, y in enumerate(hist.yvals):
                for i, x in enumerate(hist.xvals):
                    hist.data[j][i] = self.fit(xval=x, yval=y, sample=True,
                                               **kwargs)

        if self._xcons:
            del self.model.con['__xscan__']
        if self._ycons:
            del self.model.con['__yscan__']
        self.model.par = par


class Profiler1D:
    """Compute the profile likelihood on a one-dimensional space.

    Note:
      When the profile likelihood function is to be computed on a 1D grid the
      stability of the optimisations can be significantly improved by using
      the best-fit parameters of neighbouring grid points as starting points for
      local optimisations. The :py:class:`~myFitter.Profiler1D` and
      :py:class:`~myFitter.Profiler2D` classes implement this method for one and
      two-dimensional scans.

      For difficult optimisation problems you may need rather detailed control
      over the procedure for finding the global minima. In this case you can
      derive a class from Profiler1D and override the
      :py:meth:`~myFitter.Profiler1D.fit` method.
      
      The variable in which the profile likelihood is computed can be an
      arbitrary dependent quantity (not necessarily a model parameter). For a
      dependent quantity which is not a model parameter a constraint is
      (temporarily) added to the model. The name of this constraint is
      ``'__xscan__'``. You should therefore avoid using this name for your own
      constraints.

    Args:
      xvar (str): Name of the variable on the x-axis. It must
        be the name of a dependent quantity in the model (not necessarily
        a parameter).
      fitter (Fitter or None, optional): The Fitter object to use for global
        minimisations. The model whose likelihood is profiled will be
        ``fitter.model``. Defaults to ``None``, in which case the `model`
        argument must be given.
      model (Model or None, optional): The model whose likelihood function
        shall be profiled. By specifying the `model` keyword instead of `fitter`
        you disable global optimisations by Fitter objects. Use this only if 
        local optimisations reliably find the global minimum (e.g. because your
        likelihood function is convex). Defaults to ``None``, in which case
        `fitter` must be specified.
      xscale (float, optional): Scale for the constraint associated
        with the scan variable. (See :py:class:`~myFitter.Constraint` for
        details.) This argument is irrelevant if the scan variable is a model
        parameter. nsamples (int, optional): Number of sample points to
        generate for each global optimisation. Defaults to 100.
      nsamples (int, optional): Number of points to sample for each global
        minimisation. Defaults to 100.
      tries (int, optional): Number of local minimisations to try for finding
        the global minimum. Defaults to 5.
      log (output stream or None, optional): Output stream for log messages.
        Defaults to ``None``, in which case no messages are written.
      useneighbors (bool, optional): Whether to use best-fit parameters of
        neighbouring cells as starting points. Defaults to ``True``.

    """
    def __init__(self, xvar, fitter=None, model=None, xscale=1.0,
                 nsamples=100, tries=5, log=None, useneighbors=True):
        self.xvar = xvar
        self.fitter = fitter
        if model is None:
            if fitter is not None:
                self.model = fitter.model
            else:
                raise ValueError('Specify either `fitter` or `model` keyword.')
        else:
            if fitter is not None:
                raise ValueError('Specify either `fitter` or `model` keyword.')
            else:
                self.model = model
        self._xcons = self.xvar not in self.model.par.keys()
        self._xscale = xscale

        self.nsamples = nsamples
        self.tries = tries
        self.log = log
        self.useneighbors = useneighbors

    def fix(self, x):
        """Fix the variable of the profile likelihood.

        Note:
          If the variable of the profile likelihood is a model parameter it is
          fixed with :py:meth:`myFitter.Model.fix`. Otherwise the ``bound``
          member of the corresponding constraint is adjusted. You will
          typically use this function in a derived class when you override the
          :py:meth:`~myFitter.Profiler1D.fit` method.

        Args:
          x (float): Values of the variables.

        """
        if self._xcons:
            self.model.con['__xscan__'].bound = x
        else:
            self.model.fix({self.xvar : x})

    def fit(self, xval=None, start=[], sample=True,
            nsamples=100, tries=5, log=None, **kwargs):
        """Minimise the chi-square for a single grid point.

        Note:
          Derived classes can override this function to implement specific
          methods for global optimisation. The function should fix the scan
          variable to `xval` (using :py:meth:`~myFitter.Profiler1D.fix`), do
          local optimisations with all the starting points listed in `start`, and
          perform an additional global optimisation if and only if `sample` is
          ``True``. The default behaviour is to randomly sample a number of
          points set by the :py:attr:`~myFitter.Profiler1D.nsamples` attribute
          and then try a number of local optimisations set by
          :py:attr:`~myFitter.Profiler1D.tries`.

        Args:
          xval (float): Value of the scan variable.
          start (list of dicts, optional): Starting points for local
            optimisations. The dictionaries will hold values for the model
            parameters. Defaults to ``[]``.
          sample (bool, optional): Indicates if an additional global
            optimisation (i.e. random sampling followed by several local
            optimisations) is required. Defaults to ``True``.
          nsamples (int, optional): Number of points to sample for each global
            minimisation. Defaults to 100.
          tries (int, optional): Number of local minimisations to try for finding
            the global minimum. Defaults to 5.
          log (output stream or None, optional): Output stream for log messages.
            Defaults to ``None``, in which case no messages are written.
          **kwargs (optional): additional keyword arguments. The default
            implementation passes them to the :py:meth:`myFitter.Fitter.fit`
            method.

        """
        if log is not None:
            if sample:
                log.write('pass 1, x='+repr(xval)+'\n')
            else:
                log.write('pass 2, x='+repr(xval)+'\n')
            log.flush()
        self.fix(xval)
        if sample:
            self.fitter.clear()
            self.fitter.sample(nsamples)
        else:
            tries = 0
        return self.fitter.fit(tries=tries, start=start, **kwargs)

    def scan(self, hist, **kwargs):
        """Compute the profile likelihood on a specified grid.

        Args:
          hist (Histogram1D): A 1D histogram whose
            :py:attr:`~myFitter.Histogram1D.xvals`
            member specifies the grid on which the profile likelihood is to be
            computed. After a successful scan the data values of `hist` will be
            :py:class:`~myFitter.FitResult` objects holding information about
            the fit which yielded the smallest chi-square value.
          nsamples (int, optional): Number of points to sample for each global
            minimisation. Default value set by constructor.
          tries (int, optional): Number of local minimisations to try for finding
            the global minimum. Defaults value set by constructor.
          log (output stream or None, optional): Output stream for log messages.
            If ``None`` no messages are written. Default value set in
            constructor.
          useneighbors (bool or None, optional): Whether to use best-fit
            parameters of neighbouring cells as starting points. Default
            value is set in constructor.
          **kwargs (optional): Any additional keyword arguments are passed to
            the :py:meth:`~myFitter.Profiler1D.fit` method.

        """
        if 'nsamples' not in kwargs:
            kwargs['nsamples'] = self.nsamples
        if 'tries' not in kwargs:
            kwargs['tries'] = self.tries
        if 'log' not in kwargs:
            kwargs['log'] = self.log

        useneighbors = kwargs.pop('useneighbors', self.useneighbors)

        par = self.model.par
        self.model.par = copy.deepcopy(par)
        if self._xcons:
            self.model.addConstraint(self.xvar, '=', 0.0,
                                     scale=self._xscale, 
                                     active=True,
                                     name='__xscan__')

        if useneighbors:
            for i, x in enumerate(hist.xvals):
                sp = []
                if i > 0:
                    sp.append(hist.data[i-1].bestFitParameters)
                r = self.fit(xval=x, start=sp, sample=True, **kwargs)
                hist.data[i] = min(hist.data[i], r)

            if len(hist.xvals) > 1:
                for i, x in reversed(list(enumerate(hist.xvals))):
                    sp = []
                    if i < len(hist.xvals)-1:
                        sp.append(hist.data[i+1].bestFitParameters)
                    if sp:
                        r = self.fit(xval=x, start=sp, sample=False, **kwargs)
                        hist.data[i] = min(hist.data[i], r)
        else:
            for i, x in enumerate(hist.xvals):
                hist.data[i] = self.fit(xval=x, sample=True, **kwargs)

        if self._xcons:
            del self.model.con['__xscan__']
        self.model.par = par

