import unittest
import numpy
from myFitter import *


def eq(a, b, rprec=1e-10, aprec=1e-10):
    d = abs(a-b)
    if d < aprec or a == b == 0.0:
        return True
    else:
        return 2.0*abs(a-b) < rprec*(abs(a)+abs(b))

class LinearModel(Model):
    def __init__(self, mat, vec):
        Model.__init__(self)

        self.addParameter('P1')
        self.addParameter('P2')
        self.addParameter('P3')
        self.addParameter('P4')

        self._mat = numpy.array(mat)
        self._vec = numpy.array(vec)

    def getDependentQuantities(self, p):
        q = Model.getDependentQuantities(self, p)
        pvec = numpy.array([q['P1'], q['P2'], q['P3'], q['P4']])
        ovec = self._vec + numpy.dot(self._mat, pvec)
        for i, o in enumerate(['O1', 'O2', 'O3', 'O4', 'O5', 'O6', 'O7']):
            q[o] = ovec[i]
        return q


class ToysTest(unittest.TestCase):
    def setUp(self):
        numpy.random.seed(1234)

    def test_nested(self):
        fmod = LinearModel([[+0.3, -1.2, +0.0, -0.7],
                            [+0.4, +0.6, -0.9, +1.0],
                            [-1.5, -0.1, -0.1, -0.5],
                            [+0.0, +0.0, -0.6, +0.2],
                            [+0.4, -0.4, -1.1, +0.7],
                            [-0.3, -0.6, +0.2, +0.9],
                            [-0.1, +0.9, -1.0, +0.0]],
                           [+0.5, +0.5, -0.4, -0.8, -0.9, +0.2, +0.0])
        q = fmod.getDependentQuantities(
            {'P1':-1.0, 'P2':-2.0, 'P3':+0.4, 'P4':+0.9})

        fmod.addGaussianCSC('O1', q['O1']+0.3, 0.2)
        fmod.addGaussianCSC('O2', q['O2']-0.1, 0.2)
        fmod.addGaussianCSC('O4', q['O4']-0.5, 0.3)
        fmod.addGaussianCSC('O6', q['O6']+0.2, 0.1)
        fmod.addCorrelatedGaussianCSC(
            'O357',
            [('O3', q['O3']-0.5, 0.2),
             ('O5', q['O5']+0.1, 0.3),
             ('O7', q['O7']-0.2, 0.2)],
            [('O3', 'O5', -0.1),
             ('O3', 'O7', -0.5),
             ('O5', 'O7', +0.3)])

        cmod = fmod.copy()
        cmod.fix({'P1':-0.7, 'P2':-2.0})


        ffitter = Fitter(fmod)
        ffitter.sample(100)
        cfitter = Fitter(cmod)
        cfitter.sample(100)

        ffit = ffitter.fit()
        cfit = cfitter.fit()

        dchisq = cfit.minChiSquare-ffit.minChiSquare
        self.assertTrue(eq(dchisq, 3.23124, rprec=1e-4))

        integ = ToySimulator(cmod, cfit.bestFitParameters,
                             fmod, ffit.bestFitParameters,
                             dchisq=dchisq)
        self.assertTrue(integ._lam0 is None)
        result = GVar(integ(lambda x: 1.0, neval=1000, nitn=5))
        self.assertTrue(eq(result.mean, 1.0, rprec=2.0*result.sdev))

        fmod.fix({'P3':0.4})
        ffitter.clear()
        ffitter.sample(100)
        ffit = ffitter.fit()
        dchisq = cfit.minChiSquare-ffit.minChiSquare
        self.assertTrue(eq(dchisq, 1.84616101198, rprec=1e-4))

        integ = ToySimulator(cmod, cfit.bestFitParameters,
                             fmod, ffit.bestFitParameters,
                             dchisq=dchisq)
        self.assertTrue(integ._lam0 is not None)
        result = GVar(integ(lambda x: 1.0, neval=1000, nitn=5))
        self.assertTrue(eq(result.mean, 1.0, rprec=2.0*result.sdev))
